from dataclasses import dataclass
from typing import (AnyStr, Tuple, List, Dict)

from msb.cipher import Cipher
from msb.db.constants import (DEFAULT_QUERY_OFFSET, DEFAULT_QUERY_LIMIT)


@dataclass
class SearchParameter:
	# this is what goes to the search
	_model_filters: Dict
	_fields: List
	_order_by: AnyStr
	_limit: int
	_offset: int
	_order_field: str
	# this is what we recieve from payload
	_raw_filters: Dict

	# this is what we is declared beforehand
	_model_filter_map: Dict
	_auto_decrypt_filters: List

	def get_query(self, for_model=None):
		try:
			self._init_model_filters(model=for_model)
			_filter_list, _filter_dict = self.compound_filters, self.filters
			_query_fields, _query_properties = self._extract_query_fields_and_properties(for_model)

			query_set = for_model.objects.from_all

			if len(_filter_list) > 0:
				query_set = query_set.filter(*_filter_list, **_filter_dict)
			else:
				query_set = query_set.filter(**_filter_dict)

			if len(_query_fields) > 0:
				query_set = query_set.only(*_query_fields)

			if len(_query_properties) > 0:
				for_model.add_query_properties(*_query_properties)

			if self.order_field:
				query_set = query_set.order_by(self.order_field)
			return query_set
		except Exception as e:
			raise e

	def _init_model_filters(self, model=None):
		for key, value in self._raw_filters.items():
			value = self.__parse_value(key=key, value=value, model=model)

			if key in self.model_filter_map.keys():
				self.add_filters(self.model_filter_map[key], value)
			elif hasattr(model, key):
				self.add_filters(key, value)

	def _extract_query_fields_and_properties(self, model=None):
		_model_fields, _model_properties = [], []
		if self.fields:
			for f in self.fields:
				if isinstance(getattr(model, f, None), property):
					_model_properties.append(f)
				else:
					_model_fields.append(f)

		return _model_fields, _model_properties

	def __parse_value(self, key, value, model):
		if key in self.auto_decrypt_fields_list(model):
			return [Cipher.decipher(i) for i in value] if isinstance(value, list) else Cipher.decipher(value)
		return value

	def __init__(self, **kwargs):
		self._model_filters = dict()
		self._fields = kwargs.get('fields', [])
		self._order_by = kwargs.get('order_by', 'ASC')
		self._limit = kwargs.get('limit', DEFAULT_QUERY_LIMIT)
		self._offset = kwargs.get('offset', DEFAULT_QUERY_OFFSET)
		self._order_field = kwargs.get('order_field',None)
		self._raw_filters = kwargs.get('filters', {})

	def add_filters(self, key: str, value):
		if key and value not in ['', None, []]:
			self._model_filters[key] = value

	def auto_decrypt_fields_list(self, model):
		_secure_field_list = self.auto_decrypt_filters
		_secure_field_list.append(model._meta.pk.attname)
		_private_fields = getattr(model, "_private_fields", [])
		if isinstance(_private_fields, list):
			_secure_field_list.extend(_private_fields)
		return _secure_field_list

	@property
	def compound_filters(self) -> List | Tuple:
		return []

	@property
	def model_filter_map(self) -> dict:
		return getattr(self, "_model_filter_map", dict())

	@property
	def auto_decrypt_filters(self) -> list:
		return getattr(self, "_auto_decrypt_filters", list())

	@property
	def filters(self) -> dict:
		return self._model_filters

	@property
	def fields(self) -> list:
		return self._fields if isinstance(self._fields, list) and len(self._fields) > 0 else None

	@property
	def order_by(self) -> str:
		return self._order_by

	@property
	def limit(self) -> int:
		try:
			return int(self._limit) if int(self._limit) > 0 else DEFAULT_QUERY_LIMIT
		except Exception:
			return 0

	@property
	def offset(self) -> int:
		try:
			return int(self._offset) if int(self._offset) >= 0 else DEFAULT_QUERY_OFFSET
		except Exception:
			return 0

	@property
	def order_field(self) -> str:
		if self._order_field is not None:
			return f"-{self._order_field}" if str(self.order_by).upper() != 'ASC' else self._order_field
		return None


@dataclass
class SearchParameterRequest:
	fields: List
	filters: Dict
	limit: int
	offset: int
	order_field: str | None
	order_by: str | None

	def __init__(self, **kwargs):
		self.fields = kwargs.get('fields') or []
		self.filters = kwargs.get('filters') or dict()
		self.limit = kwargs.get('limit') or DEFAULT_QUERY_LIMIT
		self.offset = kwargs.get('offset') or DEFAULT_QUERY_OFFSET
		self.order_field = kwargs.get('order_field')
		self.order_by = kwargs.get('order_by')



class AppException(Exception):
	_message: str = ''
	_description: str = ''
	_errors: list = []
	_code: int = 600

	def __init__(self, message: str = None, desc: str = None, errors: list = None):
		self._description = f"\nDesc :{desc}" if isinstance(desc, str) else ""
		self._message = message if isinstance(message, str) else self._message
		super().__init__(f"{self._message}{self._description}")
		self._errors = errors if isinstance(errors, dict) else list()

	@property
	def message(self):
		return self._message

	@property
	def description(self):
		return self._description

	@property
	def errors(self):
		return self._errors

	@property
	def code(self):
		return self._code

	def log(self):
		import logging
		logging.exception(self)


class ApiException(AppException):
	def __init__(self, message: str = None, desc: str = None, errors: dict = None):
		super().__init__(message=message, errors=errors, desc=desc)


class CrudApiException(ApiException):
	_message: str = "Failed to process the {resource} data."

	def __init__(self, resource: str, desc: str = None, errors: dict = None):
		message = self._message.format(resource=(resource or "requested"))
		super().__init__(message=message, errors=errors, desc=desc)



from functools import cached_property

from rest_framework.permissions import BasePermission

from .constants import RoleConst
from .exceptions import MsbAuthExceptions
from .users import TokenUser


class Permissions:
	"""
		Default Permissions
	"""

	class Authenticated(BasePermission):
		def is_authenticated(self, request) -> bool:
			return request.user and request.user.is_authenticated

		def has_permission(self, request, view) -> bool:
			return self.is_authenticated(request)

	class IntraService(Authenticated):

		def has_permission(self, request, view) -> bool:
			return bool(self.is_authenticated(request) and request.user.is_intra_service_requester)

	class AuthenticatedUser(Authenticated):

		allowed_role_ids: list = RoleConst.ALL_USER_ROLE_IDS
		admin_is_allowed_by_default: bool = True

		@cached_property
		def __get_allowed_role_ids(self) -> list:
			allowed_ids: list = self.allowed_role_ids if isinstance(self.allowed_role_ids, list) else []
			if self.admin_is_allowed_by_default:
				allowed_ids.append(RoleConst.ADMIN_ROLE_ID)
			return allowed_ids

		def has_role_permission(self, user: TokenUser) -> bool:
			if len(self.__get_allowed_role_ids) > 0:
				return any([user.has_role(role=role_id) for role_id in self.__get_allowed_role_ids])
			return True

		def has_permission(self, request, view) -> bool:
			is_authenticated = self.is_authenticated(request)
			has_role_permission = self.has_role_permission(request.user)
			return is_authenticated and has_role_permission

		def has_object_permission(self, request, view, obj):
			return True

		@classmethod
		def check_role_permission(cls, *args, **kwargs):
			return cls()

		def check(self, request, view, obj=None) -> bool:
			if not self.has_permission(request, view, ) and self.has_object_permission(request, view, obj):
				raise MsbAuthExceptions.UnauthorizedAccess
			return True

	class Management(AuthenticatedUser):

		def has_permission(self, request, view) -> bool:
			return bool(super().has_permission(request, view) and request.user.has_management_access)


	class ManagementOrIntraService(AuthenticatedUser):
		def has_permission(self, request, view) -> bool:
			_user: TokenUser = request.user
			return bool(
				super().has_permission(request, view) and
				(_user.has_management_access or _user.is_intra_service_requester)
			)

	"""
		Role Permissions
	"""

	class AdminRole(Management):

		def has_permission(self, request, view) -> bool:
			return bool(super().has_permission(request, view) and request.user.is_admin)

	"""
		Manager Role Permissions
	"""

	class ManagerRole(Management):
		allowed_role_ids = RoleConst.ALL_MANAGER_ROLE_IDS

	class ResourceManagerRole(ManagerRole):
		allowed_role_ids = [RoleConst.RESOURCE_MANAGER_ROLE_ID]

	class FinanceManagerRole(Management):
		allowed_role_ids = [RoleConst.FINANCE_MANAGER_ROLE_ID]

	class HrManagerRole(Management):
		allowed_role_ids = [RoleConst.HR_MANAGER_ROLE_ID]

	class OperationsManagerRole(Management):
		allowed_role_ids = [RoleConst.OPERATIONS_MANAGER_ROLE_ID]

	class ItManagerRole(Management):
		allowed_role_ids = [RoleConst.OPERATIONS_MANAGER_ROLE_ID]

		def has_permission(self, request, view) -> bool:
			_user: TokenUser = request.user
			return bool(super().has_permission(request, view) and _user.has_role(RoleConst.IT_MANAGER_ROLE_ID))

	"""
		Employee Role Permissions
	"""

	class EmployeeRole(AuthenticatedUser):
		allowed_role_ids = RoleConst.ALL_MANAGER_AND_EMPLOYEE_ROLE_IDS

	class HrEmployeeRole(Management):
		allowed_role_ids = [RoleConst.HR_EMPLOYEE_ROLE_ID, RoleConst.HR_MANAGER_ROLE_ID]

	class FinanceEmployeeRole(Management):
		allowed_role_ids = [RoleConst.FINANCE_MANAGER_ROLE_ID, RoleConst.FINANCE_EMPLOYEE_ROLE_ID]

	class OperationsEmployeeRole(Management):

		allowed_role_ids = [RoleConst.OPERATIONS_MANAGER_ROLE_ID, RoleConst.OPERATIONS_EMPLOYEE_ROLE_ID]

	class ItEmployeeRole(Management):

		allowed_role_ids = [RoleConst.IT_MANAGER_ROLE_ID, RoleConst.IT_EMPLOYEE_ROLE_ID]

	class ApplicationQaRole(Management):

		allowed_role_ids = [RoleConst.APPLICATION_QA_ROLE_ID, RoleConst.ADMIN_ROLE_ID]

	class GuestRole(AuthenticatedUser):

		def has_permission(self, request, view) -> bool:
			_user: TokenUser = request.user
			return bool(super().has_permission(request, view) and _user.has_role(RoleConst.GUEST_ROLE_ID))

	class ContractorRole(AuthenticatedUser):

		def has_permission(self, request, view) -> bool:
			_user: TokenUser = request.user
			return bool(super().has_permission(request, view) and _user.has_role(RoleConst.CONTRACTOR_ROLE_ID))



from django.conf import settings
from rest_framework_simplejwt.authentication import JWTTokenUserAuthentication

from msb.auth.users import TokenUser
from msb.http import RequestWrapper
from .exceptions import MsbAuthExceptions


class JwtTokenValidator(JWTTokenUserAuthentication):
	"""
	This class overrides the "JWTTokenUserAuthentication", so that we can implement
	custom authentication rules over the default system

	"""

	_request: RequestWrapper = None

	def _validate_token_owner(self, token_user: TokenUser = None) -> bool:
		if settings.MSB_JWT_VERIFY_OWNER and not token_user.is_intra_service_requester:
			if self._request.ip in settings.MSB_JWT_TRUSTED_OWNERS:
				return True
			return f"{self._request.user_agent}@{self._request.ip}" == token_user.owner
		return True

	def _validate_subscription(self, token_user: TokenUser = None) -> bool:
		if settings.MSB_JWT_VERIFY_SUBSCRIPTIONS:
			_request_path = self._request.path.strip("/").split('/')[0]
			return token_user.has_access_to(path=_request_path)
		return True

	def _validate_environment(self, token_user: TokenUser = None) -> bool:
		if settings.MSB_JWT_VERIFY_ENVIRONMENT:
			return token_user.env.casefold() == settings.ENVIRONMENT.casefold()
		return True

	def authenticate(self, request):
		try:
			self._request = RequestWrapper(request=request)

			# authenticate the request from parent class
			if auth_result := super().authenticate(request=request):
				token_user: TokenUser = auth_result[0]

				# check if the token_user is authenticated
				if not token_user.is_authenticated:
					raise MsbAuthExceptions.AuthenticationFailed

				# check if the token belongs to the same environment as the current one
				if not self._validate_environment(token_user=token_user):
					raise MsbAuthExceptions.InvalidTokenEnvironment

				# check if the token owner is valid
				if not self._validate_token_owner(token_user=token_user):
					raise MsbAuthExceptions.InvalidTokenOwner

				# check if the user is subscribed to the service he is trying to access
				if not self._validate_subscription(token_user=token_user):
					raise MsbAuthExceptions.InvalidSubscriptionAccess

				# update the token user properties
				token_user.set_validation_status(True)

			# return the authentication result
			return auth_result
		except Exception as e:
			raise e
			return None



from functools import wraps

from .utils import (validate_inp_payload, validate_inp_parameters, Request, ValidationSchema)


def validation_schema_wrapper(klass):
	def get(key: str, default=None) -> ValidationSchema | None:
		if hasattr(klass, key) and isinstance((schema := getattr(klass, key)), ValidationSchema):
			return schema
		return default

	setattr(klass, "get", get)
	return klass


def api_inputs(payload_schema: ValidationSchema = None, param_schema: ValidationSchema = None, **opt):
	allow_unknown = opt.get("unknown") == True
	bulk_inp = opt.get("bulk_inp") == True

	def outer_func(_func):
		@wraps(_func)
		def inner_func(cls, request: Request, *args, **kwargs):

			"""
			 validate parameter data like from URL/Query string if parameter validation rules are defined
			"""
			if isinstance(param_schema, ValidationSchema):
				validate_inp_parameters(param_schema, kwargs, allow_unknown=True)

			"""
			validate payload data if payload rules are defined
			"""
			if isinstance(payload_schema, ValidationSchema):
				validate_inp_payload(payload_schema, request, allow_unknown, bulk_inp)



			cls._api_input_is_validated = True
			return _func(cls, *args, **dict(request=request, **kwargs))

		return inner_func

	return outer_func



DEFAULT_FIXTURES_DIR = "base"
TEST_FIXTURES_DIR = "test"

DEFAULT_USER_AGENT = 'Test User Agent 1.0'

CONTENT_TYPE_FIELD = "Content-Type"
CONTENT_TYPE_JSON = "application/json"
USER_AGENT_FIELD_NAME = 'HTTP_USER_AGENT'

DEFAULT_TEST_DATABASES = '__all__'
DEFAULT_TEST_SERVER_PORT = 8000

CONFIG_USER_AGENT_NAME = 'USER_AGENT_STRING'
CONFIG_AUTH_ENDPOINT = 'AUTH_ENDPOINT'

ACCESS_TOKEN_NAME = "access"
REFRESH_TOKEN_NAME = "refresh"



import datetime
from random import randint

from msb.datetime import (DEFAULT_DATE_FORMAT, DEFAULT_TIME_FORMAT, current_date, current_timestamp)

_random_names = {
	"female": ["Olivia", "Emma", "Charlotte", "Amelia", "Ava", "Sophia", "Isabella", "Mia", ],
	"male": ["Kai", "Zion", "Jayden", "Luca", "Ezra", "Kayden", "Quinn", "Rowan", ],
	"surnames": ["Anand", "Laghari", "Patel", "Reddy", "Acharya", "Agarwal", "Khatri", "Ahuja", ]

}


def random_name(gender: str = "M|F|None", surnames: bool = False) -> str:
	if surnames:
		_surnames = _random_names.get("surnames")
		return _surnames[randint(0, len(_surnames) - 1)]
	else:
		_female_names: list = _random_names.get("female")
		_male_names: list = _random_names.get("male")
		_all_names: list = [*_female_names, *_male_names]
		name_set = (_female_names if gender == "F" else _male_names) if gender in ["M", "F"] else _all_names
		return name_set[randint(0, len(name_set) - 1)]


def random_email(domain: str = "@testing.com", gender: str = "M|F|None", surnames: bool = True) -> str:
	return f"{'.'.join(random_name(gender=gender, surnames=surnames).split(' '))}{domain}"


def random_age(_min: int = 18, _max: int = 60) -> int:
	return randint(_min, _max)


def random_phone(prefix: int = 91) -> int:
	return int(f"{prefix}{randint(12345678, 99999999)}")


def random_date(start: str = "", end: str = '', fmt=DEFAULT_DATE_FORMAT):
	return current_date()


def random_time(start: str = "", end: str = '', fmt=DEFAULT_TIME_FORMAT):
	return current_timestamp()


def random_float(digits: int = 6, precision: int = 2) -> float:
	return round(randint((2 * (10 * digits)), (8 * (10 * (digits + 2)))) / 3, precision)


def random_bool():
	return randint(1, 9) < 5


class RandomUser:
	gender: str
	first_name: str
	last_name: str
	email: str

	joined_date: datetime.date
	created_at: datetime.datetime
	age: int
	salary: float
	phone: int

	roles: list
	address: dict
	status: bool

	"""
	implement this class in future
	"""

	def __init__(self, invalid=False):
		self.gender = ("M" if random_bool() else "F") if not invalid else None
		self.first_name = random_name(gender=self.gender) if not invalid else ""
		self.last_name = random_name(gender=self.gender, surnames=True) if not invalid else ""
		self.email = f"{self.first_name}@testing.com" if not invalid else ""
		self.age = random_age() if not invalid else 0
		self.salary = random_float(10000, 2) if not invalid else random_float(5, 2)
		self.phone = random_phone() if not invalid else 0
		self.roles = [1, 2, 3] if not invalid else []
		self.address = dict(temp="temp address", permanent="permanent address") if not invalid else {}
		self.joined_date = current_date() if not invalid else None
		self.created_at = current_timestamp() if not invalid else None
		self.status = random_bool() if not invalid else None

	def __selected_inputs(self, *args):
		return {k: v for k, v in self.__dict__.items() if k in args}

	@property
	def date_inputs(self):
		return self.__selected_inputs("joined_date", "created_at")

	@property
	def str_inputs(self):
		return self.__selected_inputs("gender", "first_name", "last_name", "email")

	@property
	def bin_inputs(self):
		return self.__selected_inputs("status", )

	@property
	def iter_inputs(self):
		return self.__selected_inputs("roles", "address")

	@property
	def numeric_inputs(self):
		return self.__selected_inputs("age", "salary", "phone")


__all__ = [
	"RandomUser", "random_date", "random_bool", "random_age", "random_phone",
	"random_float", "random_name", "random_time", "random_email"
]



from msb.exceptions import (ApiException)


class HttpExceptions:
	"""
	list of exceptions related to the api_gateway app
	"""

	class ResourceUnreachable(ApiException):
		_message = "Requested resource is unreachable."
		_code = 500


class ApiRequestExceptions:
	class ResourceNotFound(ApiException):
		"""
			If the requested service is not found.
		"""
		_message = "Requested resource not found."
		_code = 404

	class ResourceUnreachable(ApiException):
		_message = "Requested resource is unreachable."
		_code = 405

	class ResourceNotResponding(ApiException):
		_message = "Requested resource did not respond."
		_code = 405

	class InternalServerError(ApiException):
		_message = "Internal Server Error."
		_code = 500

	class InvalidRequestParameters(ApiException):
		_message = "Invalid request parameters for the request."
		_code = 500

	class InvalidHostOrRequestPath(ApiException):
		_message = "Invalid host or path."
		_code = 500



from abc import (abstractmethod, ABC)

from .constants import (
	DEFAULT_MIGRATION_DATABASE,
	DEFAULT_DATABASE_NAME, DEFAULT_APP_LABLES_TO_ROUTE
)


class DjangoDatabaseRouterInterface(ABC):
	OP_LIST = ('write', 'read', 'relation', 'migration')
	OP_WRITE, OP_READ, OP_RELATION, OP_MIGRATION = OP_LIST
	db_allowed_app_labels: set = DEFAULT_APP_LABLES_TO_ROUTE

	@abstractmethod
	def db_for_read(self, model, **hints):
		"""Send all read operations of the model to `__database_to_read_from`."""
		pass

	@abstractmethod
	def db_for_write(self, model, **hints) -> str:
		"""Send all write operations of the model to `__database_to_write_to`."""
		pass

	@abstractmethod
	def allow_relation(self, obj1, obj2, **hints) -> bool:
		"""Determine if relationship is allowed between two objects."""
		pass

	@abstractmethod
	def allow_migrate(self, db, app_label, model_name=None, **hints) -> bool:
		"""Ensure that the app's models get created on the right database."""
		pass


class MsbDatabaseRouter(DjangoDatabaseRouterInterface):
	db_allowed_app_labels: set = DEFAULT_APP_LABLES_TO_ROUTE
	db_to_read_from: str = DEFAULT_DATABASE_NAME
	db_to_write_to: str = DEFAULT_DATABASE_NAME
	db_to_migrate_to: set = DEFAULT_MIGRATION_DATABASE

	"""
	   Determine how to route database calls for an app's models (in this case, for an app named Example).
	   All other models will be routed to the next router in the DATABASE_ROUTERS setting if applicable,
	   or otherwise to the default database.
	   """

	def get_mapping(self, app: str, model_name: str, op: str):
		"Implement this method in child classes"
		pass

	def db_for_read(self, model, **hints):

		"""Send all read operations of the model to `__database_to_read_from`."""
		app_label = model._meta.app_label
		model_name = model.__name__.lower()
		if (_db := self.get_mapping(app=app_label, model_name=model_name, op=self.OP_WRITE)) is not None:
			return _db
		elif app_label in self.db_allowed_app_labels:
			return self.db_to_read_from
		else:
			return None

	def db_for_write(self, model, **hints) -> str:
		"""Send all write operations of the model to `__database_to_write_to`."""
		app_label = model._meta.app_label
		model_name = model.__name__.lower()
		if (_db := self.get_mapping(app=app_label, model_name=model_name, op=self.OP_WRITE)) is not None:
			return _db
		elif app_label in self.db_allowed_app_labels:
			return self.db_to_write_to
		else:
			return None

	def allow_relation(self, obj1, obj2, **hints) -> bool:
		"""Determine if relationship is allowed between two objects."""

		# Allow any relation between two models that are both in the Example app.
		if obj1._meta.app_label in self.db_allowed_app_labels or obj2._meta.app_label in self.db_allowed_app_labels:
			return True
		# No opinion if neither object is in the Example app (defer to default or other routers).
		elif self.db_to_read_from not in [obj1._meta.app_label, obj2._meta.app_label]:
			return None
		# Block relationship if one object is in the Example app and the other isn't.
		return False

	def allow_migrate(self, db, app_label, model_name=None, **hints) -> bool:
		"""Ensure that the app's models get created on the right database."""
		model_name = model_name.lower() if model_name is not None else ''
		# No opinion for the scenario as of now
		_migrate = None

		if (mapped_db := self.get_mapping(app=app_label, model_name=model_name, op=self.OP_MIGRATION)) is not None:
			_migrate = mapped_db == db

		# check if we should even try to route
		elif app_label in self.db_allowed_app_labels:
			# Allow app model to migrate on the database
			_migrate = db in self.db_to_migrate_to

		return _migrate



class ConfigObject(object):
	_config_key_name = "key"
	_config_value_name = "value"

	def __typcasted_value(self, key, val):
		try:
			return key.__class__(val)
		except Exception:
			return val

	def __parse_config_key_value(self, config_obj):
		_key = getattr(config_obj, self._config_key_name) if hasattr(config_obj, self._config_key_name) else None
		_value = getattr(config_obj, self._config_value_name) if hasattr(config_obj, self._config_value_name) else None
		return _key, _value

	def __init__(self, *config_list):
		for config in config_list:
			config_key, config_value = self.__parse_config_key_value(config_obj=config)
			if hasattr(self, config_key) and config_value is not None:
				setattr(self, config_key, config_value)



from datetime import datetime


class RuleSchema(object):
	# type of the field
	_type: str

	# checks if value contains this.
	_contains: list

	# if the field is required
	_required: bool

	# validates if the value is empty

	_empty: bool

	# minmum value (if int)
	_min: int

	# maximum value (if int)
	_max: int

	# minmum length(if str)
	_minlength: int

	# maximum length(if str)
	_maxlength: int

	# list of allowed values
	_allowed: list

	# Opposite to allowed
	_forbidden: list

	# list of particaular types
	_items: list

	# value should match, this regex
	_regex: str

	# list of sub schemas
	_a_list: list

	# dict of sub schemas
	_a_dict: dict

	# if the value can be None
	_nullable: bool

	# exclude the field name (field names in the list wont be validated)
	_excludes: list

	# all of these must be present in order for the target field to be validated.
	_dependencies: list

	def __set_min_value(self, val):
		self._min = val
		return self

	def __set_max_value(self, val):
		self._max = val
		return self

	def __set_min_length(self, val: int):
		self._minlength = val
		return self

	def __set_max_length(self, val: int):
		self._maxlength = val
		return self

	def __set_empty(self, val: bool):
		if isinstance(val, bool):
			self._empty = val
		return self

	def __set_regex(self, val: str):
		if isinstance(val, str) and len(val.strip()) > 0:
			self._regex = val
		return self

	def __set_required(self, val: bool):
		self._required = (val == True)
		return self

	def __set_contains(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._contains = val
		return self

	def __set_allowed(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._allowed = val
		return self

	def __set_forbidden(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._forbidden = val
		return self

	def __set_items(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._items = val
		return self

	def __set_a_list(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._a_list = val
		return self

	def __set_excludes(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._excludes = val
		return self

	def __set_dependencies(self, val: list):
		if isinstance(val, list) and len(val) > 0:
			self._dependencies = val
		return self

	def __set_a_dict(self, val: dict):
		if isinstance(val, dict) and len(val.keys()) > 0:
			self._a_dict = val
		return self

	def __set_nullable(self, val: bool):
		if val is not None:
			self._nullable = val
		return self

	@property
	def __schema__(self):
		defaults = {"type": self._type, }
		if hasattr(self, 'coerce') and callable(self.coerce):
			defaults['coerce'] = self.coerce

		return {**defaults, **{str(k).lstrip('_'): v for k, v in self.__dict__.items() if v not in [None]}}

	@property
	def schema(self):
		return self.__schema__

	def __init__(self, **kwargs):
		self.__set_required(kwargs.get('required'))

		self.__set_min_length(kwargs.get('min_len'))
		self.__set_max_length(kwargs.get('max_len'))

		self.__set_min_value(kwargs.get('min_value'))
		self.__set_max_value(kwargs.get('max_value'))

		self.__set_empty(kwargs.get('empty'))
		self.__set_contains(kwargs.get('contains'))
		self.__set_allowed(kwargs.get('allowed'))
		self.__set_forbidden(kwargs.get('forbidden'))
		self.__set_items(kwargs.get('items'))
		self.__set_regex(kwargs.get('regex'))

		self.__set_a_list(kwargs.get('a_list'))
		self.__set_a_dict(kwargs.get('a_dict'))
		self.__set_nullable(kwargs.get('nullable'))
		self.__set_excludes(kwargs.get('excludes'))
		self.__set_dependencies(kwargs.get('dependencies'))

	def clone(self, **kwargs):
		import copy
		_copy = copy.deepcopy(self)
		_copy.__set_required(kwargs.get('required'))
		self.__set_min_length(kwargs.get('min_len'))
		self.__set_max_length(kwargs.get('max_len'))
		self.__set_min_value(kwargs.get('min_value'))
		self.__set_max_value(kwargs.get('max_value'))
		_copy.__set_empty(kwargs.get('empty'))
		_copy.__set_contains(kwargs.get('contains'))
		_copy.__set_allowed(kwargs.get('allowed'))
		_copy.__set_forbidden(kwargs.get('forbidden'))
		_copy.__set_items(kwargs.get('items'))
		_copy.__set_regex(kwargs.get('regex'))
		_copy.__set_a_list(kwargs.get('a_list'))
		_copy.__set_a_dict(kwargs.get('a_dict'))
		_copy.__set_nullable(kwargs.get('nullable'))
		_copy.__set_excludes(kwargs.get('excludes'))
		_copy.__set_dependencies(kwargs.get('dependencies'))
		return self


class DateTimeRuleSchema(RuleSchema):
	_type = 'datetime'
	_format = '%Y-%m-%d %H:%M:%S'

	def coerce(self, date):
		return datetime.strptime(date, self._format)


class InputField(object):
	# Python Types : string
	class String(RuleSchema):
		_type = 'string'

	# Python Types : float, int, excl. bool
	class Number(RuleSchema):
		_type = 'number'

	# Python Types : int
	class Integer(RuleSchema):
		_type = 'integer'

	# Python Types : float, int, excl. bool
	class Float(RuleSchema):
		_type = 'float'

	# Python Types : bool
	class Boolean(RuleSchema):
		_type = 'boolean'

	# Python Types : collections.abc.Mapping
	class Dictionary(RuleSchema):
		_type = 'dict'

	# Python Types : collections.abc.Sequence, excl. string
	class List(RuleSchema):
		_type = 'list'

	# Python Types : set
	class Set(RuleSchema):
		_type = 'set'

	# Python Types : bytes, bytearray
	class Binary(RuleSchema):
		_type = 'binary'

	# Python Types : datetime.date
	class Date(DateTimeRuleSchema):
		_type = 'date'
		_format = '%Y-%m-%d'

	# Python Types : datetime.datetime
	class Datetime(DateTimeRuleSchema):
		pass

	# Python Types : datetime.time
	class Time(DateTimeRuleSchema):
		_format = '%H:%M:%S'


class ValidationSchema(object):

	def __init__(self, **kwargs):
		for k, v in kwargs.items():
			setattr(self, k, v)

	@property
	def parsed_schema(self):
		_schema = {k: v.__schema__ if hasattr(v, "__schema__") else v.__dict__ for k, v in self.__dict__.items()}
		return _schema

	def get(self, key: str, default=None):
		try:
			if hasattr(self, key) and isinstance((schema := getattr(self, key)), RuleSchema):
				return schema.__schema__ if hasattr(schema, "__schema__") else schema.__dict__
			return default
		except Exception as e:
			return None



from msb.exceptions import ApiException


class MsbAuthExceptions:
	class AuthenticationFailed(ApiException):
		"""
			general exception that user needs to see
		"""
		_message = "Authentication failed."
		_code = 640

	class InvalidRequestParameters(ApiException):
		"""
		If failed to build the auth data
		"""
		_message = "Invalid request parameters."
		_code = 641

	class InvalidTokenOwner(ApiException):
		_message = "Token Owner has changed."
		_code = 641

	class InvalidTokenEnvironment(ApiException):
		_message = "Access token belongs to another environment."
		_code = 641

	class InvalidSubscriptionAccess(ApiException):
		_message = "You are not subscribed to this service."
		_code = 641

	class UnauthorizedAccess(ApiException):
		_message = "You are not authorized to perform this action."
		_code = 641

	class UnauthorizedAccessToUserData(ApiException):
		_message = "You are not authorized to access this user data."
		_code = 641



from msb.exceptions import CrudApiException, AppException


class ApiServiceExceptions:
	class InvalidDatabaseModel(CrudApiException):
		_message = "Invalid database model for {resource} in services."

	class InvalidPk(CrudApiException):
		_message = "Primary Identifier Field for {resource} can't be empty."

	class InvalidDataForCreateOperation(CrudApiException):
		_message = "Invalid data for {resource} creation."

	class CreateOperationFailed(CrudApiException):
		_message = "Failed to create {resource}."

	class RetrieveOperationFailed(CrudApiException):
		_message = "Failed to retrieve {resource} details."

	class ResourseDoesNotExists(CrudApiException):
		_message = "Requested {resource} does not exist."

	class ListOperationFailed(CrudApiException):
		_message = "Failed to retrieve {resource} list."

	class UpdateOperationFailed(CrudApiException):
		_message = "Failed to update {resource}."

	class BulkUpdateOperationFailed(CrudApiException):
		_message = "Failed to update {resource} in bulk."

	class DeleteOperationFailed(CrudApiException):
		_message = "Failed to delete {resource}."

	class BulkDeleteOperationFailed(CrudApiException):
		_message = "Failed to delete {resource} in bulk."

	class SearchOperationFailed(CrudApiException):
		_message = "Failed to search {resource}."

	class InvalidSearchField(CrudApiException):
		_message = "Invalid search field for {resource}."

	class DuplicateEntry(CrudApiException):
		_message = "{resource} already exists."


class NotificationServiceExceptions:
	class InvalidNotificationData(AppException):
		_message = "Notification data should be of type NotificationData."

	class NotificationSendFailed(AppException):
		_message = "failed to send notification."



def singleton(class_):
	instances = {}

	def getinstance(*args, **kwargs):
		if class_ not in instances:
			instances[class_] = class_(*args, **kwargs)
		return instances[class_]

	return getinstance


class Singleton(type):
	_instances = {}

	def __call__(cls, *args, **kwargs):
		if cls not in cls._instances:
			cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
		return cls._instances[cls]


class VirtualDataClass:
	__fields: dict = dict()
	_request_data: dict = dict()
	_resource_function = None

	@property
	def fields(self) -> dict:
		return self.__fields

	def get(self, key, cast_to: any = str) -> any:
		self._fetch_virtual_data()
		return cast_to(self.__fields.get(key))

	def _fetch_virtual_data(self):
		if len(self.__fields.keys()) > 0:
			return
		if callable(self._resource_function):
			self.__fields = self._resource_function(data=self._request_data)
